;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~        (C) Copyright raphi Software 2004 - All Rights Reserved      ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~ Author : Raphal PASQUIER                                           ~
;~ Email  : raphael.pasquier@neuf.fr                                   ~
;~ Version: 0.0.0.2                                                    ~
;~ Date   : 13/07/2005                                                 ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~                                                                     ~
;~    exemple d' utilisation de yasm et du C avec gcc sur GNU/linux    ~
;~                                                                     ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~                                                                     ~
;~  yasm -m amd64 -f elf software.asm -o software.o                    ~
;~                                                                     ~
;~  gcc -m64 software.o main.c -o main                                 ~
;~                                                                     ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;~                                                                     ~
;~ Description : fait la somme de deux entiers longs positifs          ~
;~                                                                     ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BITS 64
	
GLOBAL divHalfB_R_x86_64_bini


; en C => long int divHalfB_R_x86_64_bini( Lii* a , long int D ) ;


; J' utilise l'algorithme de Montgomery et Torbjorn
; article "division by invariant integers using multiplication"
; Avec les notations de l'article, Je prends :
; N = 64
; l = adapt au diviseur d
; ainsi le diviseur D doit vrifier :
; 0 < D < 2^32
;
; la Base est Bh = 2^32 , 


divHalfB_R_x86_64_bini:

	mov rcx, [rdi]  ; rcx = signe de a
	test rcx, rcx
	jnz non_nul
	ret 
	
non_nul:
	dec rsi  
	test rsi, rsi   ;    D = 1  =>  on sort
	jnz non_un
	ret
	
	; *** Utilisation des registres ***
	; rdx : rax  = produit
	; rbx        = tempon
	; rcx        = l
	; rdi        = adresse de a mais pas dans la boucle
	; rsi        = D = diviseur
	; r8         = adresse de mantisse ou reste
	; r9         = tempon
	; r10        = m'
	; r11        = quotient

non_un:
	inc rsi
	
	push rbp
	mov rbp, rsp
	mov [rbp-8] , rbx ; on sauvegarde rbx dans la zone rouge
	
	; ***  on calcule l' entier m'   ***
	; ** on commence par l
	bsr rcx, rsi     
	bsf rdx, rsi
	cmp rcx, rdx
	jz non_inc
	inc rcx          
non_inc:
	; rcx = l = [log2(rsi)] + 1
	mov r10, 1      ; r10 <- 1  
	shl r10, cl     ; r10 = 2^l
	sub r10, rsi    ; r10 = 2^l - d
	mov rdx, r10    ; rdx <- r10
	dec rcx          ; r9 = l - 1
	xor rax, rax    ; rax = 0
	div rsi		; rax = quotient = m' - 1
	inc rax
	mov r10, rax    ; r10 <- m'  (cf article)

	; ***  initialisation   ***
	mov r9, [rdi+32]  ; r9 = indiceMax
	shl r9, 3         ; r9 = r9 * 8
	mov r8, [rdi+16]  ; r8 = adresse de mantisse
	add r8, r9        ; r8 = adresse de mantisse[indiceMax] 


	; on calcule le reste et le quotient des 32 bits
	; de poids fort de a.mantisse[indiceMax]
	mov rax, [r8]
	shr rax,  32
	mov r11, rax
	mov r9, rax
	mul r10
	sub r11, rdx
	shr r11, 1
	add r11, rdx
	shr r11, cl    ; r11 = quotient n/D
	mov rbx, r11   ; rbx = quotient n/D
	shl rbx, 32    ; rbx = quotient * 2^32
	
	; on rcupre le reste
	mov rax, r11
	mul rsi
	sub r9, rax  ; r9 = reste
	shl r9, 32   ; r9 = reste * 2^32

	; on calcule le reste et le quotient des 32 bits
	; de poids faible de a.mantisse[indiceMax]
	mov rax, [r8]
	shl rax, 32 ;and rax,  0x00000000ffffffff
	shr rax, 32 ;and rax,  0x00000000ffffffff
	or rax, r9          ; rax = rax + rdx
	mov r9, rax
	mov r11, rax
	mul r10
	sub r11, rdx
	shr r11, 1
	add r11, rdx
	shr r11, cl    ; r11 = quotient n/D
	or rbx, r11      ; rbx = rbx + r11
	mov [r8], rbx
	; on rcupre le reste
	mov rax, r11
	mul rsi
	sub r9, rax  ; r9 = reste
	
	
	mov r8, [rdi+32]
	shl r8, 3
	
	test rbx, rbx
	jnz suitess
	mov rax, [rdi+32] ; il faut dcrmenter indiceMax
	dec rax
	mov [rdi+32], rax
	js indice_negatif  ; (indiceMax < 0 and rax ==0) => le nombre est zro
	jmp suitess

indice_negatif:
	xor rax, rax
	mov [rdi], rax  ; le nombre est zro !!!
	mov rax, r9
	mov rbx, [rbp-8]
	pop rbp
	ret
	

suitess:
	; A ce niveau r9 = reste
	;             r8 = 8 * indiceMax ancien
	mov rdi, [rdi+16]  ; ouf! on gagne le registre r8 
	
boucle:
	sub r8, 8
	js fin
	; on calcule le reste et le quotient des 32 bits
	; de poids fort de a.mantisse[i]
	mov rax, [rdi+r8]
	shr rax,  32
	
	shl r9, 32     ; r9 = reste * 2^32
	or  rax, r9    ; rax = rax + r9 
	
	mov r11, rax
	mov r9, rax
	mul r10
	sub r11, rdx
	shr r11, 1
	add r11, rdx
	shr r11, cl    ; r11 = quotient n/D
	mov rbx, r11   ; rbx = quotient n/D
	shl rbx, 32    ; rbx = quotient * 2^32
	
	; on rcupre le reste
	mov rax, r11
	mul rsi
	sub r9, rax  ; r9 = reste
	shl r9, 32   ; r9 = reste * 2^32

	; on calcule le reste et le quotient des 32 bits
	; de poids faible de a.mantisse[indiceMax]
	mov rax, [rdi+r8]
	shl rax, 32 ;and rax,  0x00000000ffffffff
	shr rax, 32 ;and rax,  0x00000000ffffffff
	or  rax, r9          ; rax = rax + r9
	mov r9, rax
	mov r11, rax
	mul r10
	sub r11, rdx
	shr r11, 1
	add r11, rdx
	shr r11, cl    ; r11 = quotient n/D
	or rbx, r11      ; rbx = rbx + r11
	mov [rdi+r8], rbx
	; on rcupre le reste
	mov rax, r11
	mul rsi
	sub r9, rax  ; r9 = reste

	jmp boucle
	
 
fin:
	mov rax, r9
	mov rbx, [rbp-8]
	pop rbp
	ret			; retour  la routine appelante



